# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: __init__.py
# Project: api
# Author: Brian Cherinka
# Created: Friday, 23rd October 2020 1:53:29 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Friday, 23rd October 2020 1:53:29 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

#
# Items to consider from the Brain
# - setting different API urls  (local, test, prod, public) - submit requests to development APIs
# - toggling different base API urls - submit requests to different APIs if available
# - tree replants for different base API urls - toggle tree replants for different APIs
# - wrapper to httpx or requests - provide convienences for submitting http requests
# - accessing API route mappings - look up available routes for a given API
# - JWT token or OAuth authentication - setting Auth precendence e.g. netrc over JWT
# - request/response compression schemes - toggle different compression algorithms

# - placeholder images - https://placeholder.com/
# - http tester - https://httpbin.org/

